//////////////////////////////////////////////////////////////////////////////////
// Engineer:       James C. Ahlstrom
// 
// Create Date:    7 December 2009
// Design Name:    Digital receiver and transmitter
// Module Name:    Receiver
// Project Name:   Transceiver
// Target Devices: Cyclone 3
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

module Receiver(
	input clock,
	input [13:0] adc_data,
	input eth_nRESET,
	input [31:0] rx_tune_phase,
	input [31:0] tx_tune_phase,
	input [7:0] rx_ctrl,
	output reg [31:0] mem_data,
	output reg [9:0] mem_write_addr,
	output reg mem_wen,
	output reg [31:0] mem_checksum, 	// must be unsigned
	output reg mem_block,				// Index of available block 0 or 1
	output [13:0] rf_ref,				// RF equal to Rx frequency for VNA
	input [15:0] vna_count				// number of VNA scan frequencies, or zero
	);
	
	// The samples can come from the receiver decimators, or from the VNA module.
	wire use_vna;
	assign use_vna = |vna_count;	// use_vna is True if vna_count != 0
	wire data_avail;
	wire signed [23:0] data_real, data_imag;
	assign data_avail = use_vna ? vna_avail : decim_avail;
	assign data_real  = use_vna ? vna_real  : decim_real;
	assign data_imag  = use_vna ? vna_imag  : decim_imag;


	// Memory blocks for ADC data
	reg [31:0] mem_data3;		// data to write
	reg [15:0] mem_save;		// odd two bytes to save from 6-byte sample
	reg [31:0] checksum;		// must be unsigned

	reg [3:0] rx_state;		// state machine for receive
	parameter sRx		= 0;
	parameter sRx1		= 1;
	parameter sRx2		= 2;
	parameter sRx3		= 3;
	parameter sRx4		= 4;
	parameter sRx5		= 5;
	parameter sRx6		= 6;
	parameter sRx7		= 7;
	parameter sRx8		= 8;
	parameter sRx9		= 9;
	parameter sRx10		= 10;
	parameter sRx11		= 11;
	
	// Tune ADC data with CORDIC
	wire signed [17:0] cordic_real, cordic_imag;
	cordic2 rx_cordic (adc_data, clock, tune_phase, cordic_real, cordic_imag, rf_ref);
	
	// Receive CIC filters followed by FIR filter
	wire decimA_avail, decimB_avail, decim_avail;
	wire signed [17:0] decimA_real, decimB_real;
	wire signed [17:0] decimA_imag, decimB_imag;
	wire signed [23:0] decim_real, decim_imag;
	CicDecimM5 #(.RRRR(8), .IBITS(18), .OBITS(18), .GBITS(15)) cic1 (
		clock, ~use_vna, cordic_real, cordic_imag,
				decimA_avail, decimA_real, decimA_imag);
	CicVarDecimM5 cic2 (rx_ctrl[5:0],
		clock, decimA_avail, decimA_real, decimA_imag,
				decimB_avail, decimB_real, decimB_imag);
	firX8R8 fir2 (clock, decimB_avail, decimB_real, decimB_imag,
			decim_avail, decim_real, decim_imag);

	// This the the VNA module.  It runs as an alternative to the above filters.
	wire vna_avail;
	wire signed [23:0] vna_real, vna_imag;
	wire [31:0] tune_phase;
	VNA m_vna (clock, eth_nRESET, cordic_real, cordic_imag, vna_count,
		rx_tune_phase, tx_tune_phase, tune_phase,
		vna_avail, vna_real, vna_imag);
	
	//reg [10:0] jim;	// Create an impulse 0, 0, 1, , 0, ...
	//always @(posedge decimB_avail)
	//	if (jim == 560)
	//		jim <= 0;
	//	else
	//		jim <= jim + 1'd1;
	//assign decimB_real = (jim == 0)? 18'd4096 : 18'd0;
	//assign decimB_imag = (jim == 0)? 18'd4096 : 18'd0;
	//assign decimB_real = 18'd4096;	// create constant x input to check sum of coef
	//assign decimB_imag = 18'd4096;

	wire [17:0] csum;	// unsigned for checksum
	assign csum = mem_data[15:0] + mem_data[31:16];

	always @(posedge clock)		// Rx state machine
	begin
		if (eth_nRESET == 0)	// we are in reset
		begin
			mem_write_addr <= 1'd0;
			mem_wen <= 0;
			rx_state <= sRx;
		end
		else case (rx_state)
		sRx:		// Wait for first sample of two samples
		begin
			if (data_avail)
			begin		// write first dword, save last two bytes LITTLE-ENDIAN
				mem_data <= {data_imag[7:0], data_real};
				mem_save <= data_imag[23:8];
				rx_state <= sRx1;
			end
		end
		sRx1:
		begin
			mem_wen <= 1'd1;
			rx_state <= sRx2;
		end
		sRx2:
		begin
			mem_wen <= 1'd0;
			checksum <= checksum + csum;
			mem_write_addr <= mem_write_addr + 1'd1;
			rx_state <= sRx3;		// wait for next sample
		end
		sRx3:		// Wait for second sample of two samples
		begin
			if (data_avail)
			begin
				mem_data <= {data_real[15:0], mem_save};
				mem_data3 <= {data_imag, data_real[23:16]};
				rx_state <= sRx4;
			end
		end
		sRx4:		// write second and third dword
		begin
			mem_wen <= 1'd1;
			rx_state <= sRx5;
		end
		sRx5:
		begin
			mem_wen <= 1'd0;
			checksum <= checksum + csum;
			rx_state <= sRx6;
		end
		sRx6:
		begin
			mem_data <= mem_data3;
			mem_write_addr <= mem_write_addr + 1'd1;
			rx_state <= sRx7;
		end
		sRx7:
		begin
			mem_wen <= 1'd1;
			rx_state <= sRx8;
		end
		sRx8:
		begin
			mem_wen <= 1'd0;
			checksum <= checksum + csum;
			rx_state <= sRx9;
		end
		sRx9:		// increment memory address
		begin
			if (mem_write_addr == Transceiver.MEM_DWORD_SIZE - 1)
			begin		// first block of memory is ready to be sent
				mem_write_addr <= mem_write_addr + 1'd1;
				mem_checksum <= checksum;
				checksum <= 1'd0;
				rx_state <= sRx10;
			end
			else if (mem_write_addr == Transceiver.MEM_DWORD_SIZE * 2 - 1)
			begin		// second block of memory is ready to be sent
				mem_write_addr <= 1'd0;
				mem_checksum <= checksum;
				checksum <= 1'd0;
				rx_state <= sRx11;
			end
			else
			begin
				mem_write_addr <= mem_write_addr + 1'd1;
				rx_state <= sRx;
			end
		end
		sRx10:
		begin
			mem_block <= 1'd0;
			rx_state <= sRx;
		end
		sRx11:
		begin
			mem_block <= 1'd1;
			rx_state <= sRx;
		end
		endcase
	end

endmodule
	
module VNA (
	input clock,
	input eth_nRESET,
	input signed [17:0] cordic_real,
	input signed [17:0] cordic_imag,
	input [15:0] vna_count,
	input [31:0] rx_tune_phase,
	input [31:0] tx_tune_phase,
	output reg [31:0] tune_phase,
	output reg vna_avail,
	output reg signed [23:0] vna_real,
	output reg signed [23:0] vna_imag
	);

	parameter CountWait = 14'd8000;
	parameter CountData = 14'd4096;
	reg [13:0] counter;
	reg [15:0] npixels;
	reg signed [29:0] sum_real, sum_imag;

	// State machine for VNA
	reg [3:0] vna_state;
	parameter sVna0		= 0;
	parameter sVna1		= 1;
	parameter sVna2		= 2;
	parameter sVna3		= 3;
	parameter sVna4		= 4;
	parameter sVna5		= 5;

	always @(posedge clock)		// VNA state machine
	begin
		if (eth_nRESET == 0 || vna_count == 0)	// we are in reset, or non-VNA
		begin
			vna_state <= sVna0;
			counter <= CountWait;
			tune_phase <= rx_tune_phase;
			npixels <= vna_count;
		end
		else case (vna_state)
		sVna0:		// Throw away data after frequency change
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
			end
			else
			begin
				vna_state <= sVna1;
				counter <= CountData;
			end
		end
		sVna1:		// Sum data
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
				sum_real <= sum_real + cordic_real;
				sum_imag <= sum_imag + cordic_imag;
			end
			else
			begin
				vna_state <= sVna2;
				counter <= CountWait;
			end
		end
		sVna2:		// Send an output datum
		begin
			vna_avail <= 1;
			vna_real <= sum_real[29:6];
			vna_imag <= sum_imag[29:6];
			vna_state <= sVna3;
		end
		sVna3:		// Increment the frequency, and repeat
		begin
			vna_avail <= 0;
			sum_real <= 1'd0;
			sum_imag <= 1'd0;
			if (npixels != 1'd1)
			begin
				npixels <= npixels - 1'd1;
				tune_phase <= tune_phase + tx_tune_phase;
				vna_state <= sVna0;
			end
			else
			begin
				npixels <= vna_count;
				tune_phase <= rx_tune_phase;
				vna_state <= sVna4;
			end
		end
		sVna4:		// Frequency reset to start, send a zero sample to mark the start
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
			end
			else
			begin
				vna_avail <= 1'd1;
				vna_real <= 1'd0;
				vna_imag <= 1'd0;
				vna_state <= sVna5;
			end
		end
		sVna5:
		begin
			vna_avail <= 1'd0;
			counter <= CountWait;
			vna_state <= sVna0;
		end
		endcase
	end
endmodule
